#ifndef TATAMI_HDF5_HPP
#define TATAMI_HDF5_HPP

/**
 * @file tatami_hdf5.hpp
 * @brief Create matrix representations for HDF5-backed data.
 */

/**
 * @namespace tatami_hdf5
 * @brief Representations for matrix data in HDF5 files.
 */
namespace tatami_hdf5 {}

#include "Hdf5CompressedSparseMatrix.hpp"
#include "Hdf5DenseMatrix.hpp"
#include "load_hdf5_matrix.hpp"
#include "write_sparse_matrix_to_hdf5.hpp"

#endif
